/***********************************************************************
**  Device driver for Oregon Micro System's PC-38
**  An Intelligent Motor Controller.
**  Linux 2.0 Operating System.
** --------------------------------------------------------------------
**  Written by:
**     Tony Denault
**     Institute for Astronomy, University of Hawaii
**     2680 Woodlawn Drive, Honolulu, HI 96822
**     Ph: (808)956-8101
**     email: denault@irtf.ifa.hawaii.edu
************************************************************************
*/

/*------------------------------
** include files
**------------------------------
*/

#include <linux/module.h>
#include <linux/version.h>

#include <linux/types.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/malloc.h>

#include <linux/pc38.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

/*------------------------------------
** device defines
**-----------------------------------
*/
#define PC38_MAJOR_NO  41   /* major device number */
#define PC38_NO         1   /* number of minor devices */

#define PC38_VERSION  0020  /* This driver's version number 0.20  */
#define DEBUG 0

/*----------------------------------------
**  Structures
**----------------------------------------
*/

/*-----------------------------
** prototype functions
**-----------------------------
*/
static int  pc38_open( struct inode * inode, struct file * file);
static void pc38_release( struct inode * inode, struct file * file);
static int  pc38_ioctl( struct inode *inode, struct file *file,
   unsigned int cmd, unsigned long arg);

/* low level functions */
static void PC38csleep( int csec );
static int PC38_puts( int minor, char *s );
static int PC38_gets( int minor, char *s, int s_len );
int PC38_clear_inbuf( int minor );
static int PC38_command( int minor, char *command, char *reply,  int r_len );

/*-----------------------------
** global variable
**-----------------------------
*/

static struct pc38_table_t pc38_table[PC38_NO] =
{
   {
      0x320,   /* base            */
   },
};

/* Define device driver entries points */
struct file_operations pc38_fops =
{
   NULL,         /* _seek    */
   NULL,         /* _read    */
   NULL,         /* _write   */
   NULL,         /* _readdir */
   NULL,         /* _select  */
   pc38_ioctl,   /* _ioctl   */
   NULL,         /* _mmap    */
   pc38_open,    /* _open    */
   pc38_release, /* _release */
};

static struct semaphore Sem = MUTEX;

/*-------------------------------------------------------------
** pc38_init() - Called at time the device driver is loaded
**-------------------------------------------------------------
*/
int pc38_init ( void )
{
   int err, 
	    minor;

	printk("Oregon Micro System PC-38 Motor Controller: major %d\n", PC38_MAJOR_NO);

   /* register device & request a major number */

   err = register_chrdev( PC38_MAJOR_NO, "pc38", &pc38_fops);
   if( err < 0 )
   {
		if(DEBUG) printk( KERN_NOTICE "pc38_init(): error %d\n", -err);
      return err;
   }

   /* Configure the device */
   for( minor=0; minor < PC38_NO; minor++ )
   {
		/* identify the IOport used by the device  */
		request_region( pc38_table[minor].base, PC38_NUM_IOPORTS, "pc38");

		printk("   PC-38 minor %d base=0x%x \n", minor, pc38_table[minor].base );
   }
	return 0;
}

/*-------------------------------------------------------------
**  Driver _open() routine. Called when process open driver.
**-------------------------------------------------------------
*/
static int pc38_open( struct inode * inode, struct file * file)
{
   if(DEBUG) printk("pc38_open(): enter\n");
   return 0;
}

/*-------------------------------------------------------------
**  Driver _release() routine. Called when process closes driver.
**-------------------------------------------------------------
*/
static void pc38_release( struct inode * inode, struct file * file)
{
   if(DEBUG) printk("pc38_release(): enter\n");
}

/*-------------------------------------------------------------
**  Driver _ioctl() routine.
**-------------------------------------------------------------
*/
static int pc38_ioctl( struct inode *inode, struct file *file,
                     unsigned int cmd, unsigned long arg)
{
   struct pc38parms_t pc38;   /*  driver's copy of user's parameters */
   int minor;

   minor = MINOR(inode->i_rdev);
   if( minor >= PC38_NO )
      return -ENODEV;

   /*--------------------------------------------------
   ** Read done flag and OR set bits into df variable 
   */
	pc38_table[minor].df |=  inb_p( pc38_table[minor].base + PC38_DONE_REG );

   /*---------------------------------------
   ** copy pc38parms_t into kernel variable
   */
   memcpy_fromfs( &pc38, (struct pc38parms_t *)arg, sizeof(struct pc38parms_t));

   switch ( cmd )
   {
      case PC38Test:
			PC38csleep( (int) pc38.l[0]);
         pc38.ret = PC38_ERR_NONE;
         break;

      case PC38GetConfig:
         pc38.ret = PC38_ERR_NONE;
			pc38.l[0] = pc38_table[minor].base;
         break;

      case PC38Inb:
         pc38.l[1] = inb_p(pc38.l[0]);
         pc38.ret = PC38_ERR_NONE;
         break;

      case PC38Outb:
         outb_p(pc38.l[1], pc38.l[0]);
         pc38.ret = PC38_ERR_NONE;
         break;

      case PC38Puts:
         down( &Sem );
		   pc38.ret = PC38_puts( minor, pc38.p[0]);
         up( &Sem );
         break;

      case PC38Gets:
         down( &Sem );
		   pc38.ret = PC38_gets( minor, pc38.p[0], pc38.l[0]);
         up( &Sem );
         break;

      case PC38Command:
         down( &Sem );
		   pc38.ret = PC38_command( minor, pc38.p[0], pc38.p[1], pc38.l[0]);
         up( &Sem );
         break;

      case PC38GetDoneFlag:
			/* no action since all call returns pc38.df */
         pc38.ret = PC38_ERR_NONE;
         break;

      case PC38ClearDoneFlag:
	      pc38_table[minor].df &= ~(0xff & pc38.l[0]);
         pc38.ret = PC38_ERR_NONE;
         break;

      case PC38GetStatus:
	      pc38.l[0] = inb_p( pc38_table[minor].base + PC38_CONTROL_REG );
	      pc38.l[1] = inb_p( pc38_table[minor].base + PC38_STATUS_REG );
         pc38.ret = PC38_ERR_NONE;
         break;

       default:
         pc38.ret = PC38_ERR_INV_FUN;
         break;
   }

   /*---------------------------------------
   ** copy pc38parms_t into user space
   */
	pc38.df = pc38_table[minor].df;
   memcpy_tofs( (struct pc38parms_t *)arg, &pc38, sizeof(struct pc38parms_t));
   return 0;
}

/*************************************************************************
*   Static function for device driver 
**************************************************************************/

/*--------------------------------------------------------------
**  PC38csleep() - a sleep function in units fo 1/100 seconds.
**                csec = 1 means sleep for 0.01 seconds.
**--------------------------------------------------------------
*/
void PC38csleep( int csec )
{
   current->state = TASK_INTERRUPTIBLE;
   current->timeout = jiffies + csec;
   schedule();
}

/*-------------------------------------------------------------------------
**  PC38_puts() - writes a string to the PC-38 using shared mem interface.
**-------------------------------------------------------------------------
*/
int PC38_puts( int minor, char *s )
{
   int cnt, ch;
   int status_reg = pc38_table[minor].base + PC38_STATUS_REG;
   int data_reg   = pc38_table[minor].base + PC38_DATA_REG;
 
   while( (ch = get_user(s++)) )
   {
      cnt = 0;
      while( (inb_p(status_reg) & PC38_TBE_BIT) == 0)
      {
         PC38csleep(1);
         if( ++cnt>5 ) 
            return PC38_ERR_TIMEOUT;
      }
      outb_p(ch, data_reg);
   }
	return PC38_ERR_NONE;
}


/*-------------------------------------------------------------------------
**  PC38_gets() - reads a string from the PC-38 into dest buf using shared 
**     mem interface. If dest buf is too small, data will be lost.
**-------------------------------------------------------------------------
*/
int PC38_gets( int minor, char *s, int s_len )
{
   int i, cnt;
   int status_reg = pc38_table[minor].base + PC38_STATUS_REG;
   int data_reg   = pc38_table[minor].base + PC38_DATA_REG;
 
   /* PC38_outp8(addr+PC38_CONTROL_REG,0); */
   for (i=0; i<s_len; i++)
   {
      cnt = 0;
      while ((inb_p(status_reg) & PC38_IBF_BIT) == 0)
      {
         PC38csleep(1);
         if( ++cnt > 5 ) 
         {
            put_user( 0, s);
            return PC38_ERR_NONE;
         }
      }
      put_user(inb_p(data_reg), s);
      s++;
   }

   put_user( 0, s);
	return PC38_ERR_NONE;
}

/*-------------------------------------------------------------------------
**  PC38_clear_inbuf() - Read from the data_reg until it is empty.
**-------------------------------------------------------------------------
*/
int PC38_clear_inbuf( int minor )
{
   int cnt;
   int status_reg = pc38_table[minor].base + PC38_STATUS_REG;
   int data_reg   = pc38_table[minor].base + PC38_DATA_REG;
 
   while( 1 )
	{
      cnt = 0;
		while ((inb_p(status_reg) & PC38_IBF_BIT) == 0)
		{
			PC38csleep(1);
			if( ++cnt > 2 ) 
				return PC38_ERR_NONE;
		}

		inb_p(data_reg);
	}

}

/*-------------------------------------------------------------------------
**  PC38_command() - user  _puts() & _gets to write/read a command/reply 
**     to the pc38. 
**-------------------------------------------------------------------------
*/
int PC38_command( int minor, char *command, char *reply,  int r_len )
{
   int rc;

   /* clear input buffer */
	PC38_clear_inbuf( minor );

   /* write command */
	if( (rc = PC38_puts( minor, command )) != PC38_ERR_NONE )
		return rc;

   /* sleep 0.05 second */
	PC38csleep( 5 );

	/* reads reply */
	rc = PC38_gets( minor, reply, r_len);

	return rc;
}


